@@ -6,13 +6,17 @@ from __future__ import print_function
 __all__ = ['fetch']
 
 import errno
+import functools
 import io
+import itertools
+import json
 import logging
 import random
 import re
 import stat
 import sys
 import tempfile
+import time
 
 from collections import OrderedDict
 
@@ -23,16 +27,23 @@ except ImportError:
 
 import portage
 portage.proxy.lazyimport.lazyimport(globals(),
+	'glob:glob',
 	'portage.package.ebuild.config:check_config_instance,config',
 	'portage.package.ebuild.doebuild:doebuild_environment,' + \
 		'_doebuild_spawn',
 	'portage.package.ebuild.prepare_build_dirs:prepare_build_dirs',
+	'portage.util:atomic_ofstream',
+	'portage.util.configparser:SafeConfigParser,read_configs,' +
+		'ConfigParserError',
+	'portage.util.install_mask:_raise_exc',
+	'portage.util._urlopen:urlopen',
 )
 
 from portage import os, selinux, shutil, _encodings, \
 	_movefile, _shell_quote, _unicode_encode
 from portage.checksum import (get_valid_checksum_keys, perform_md5, verify_all,
-	_filter_unaccelarated_hashes, _hash_filter, _apply_hash_filter)
+	_filter_unaccelarated_hashes, _hash_filter, _apply_hash_filter,
+	checksum_str)
 from portage.const import BASH_BINARY, CUSTOM_MIRRORS_FILE, \
 	GLOBAL_CONFIG_PATH
 from portage.data import portage_gid, portage_uid, secpass, userpriv_groups
@@ -253,6 +264,172 @@ _size_suffix_map = {
 	'Y' : 80,
 }
 
+
+class FlatLayout(object):
+	def get_path(self, filename):
+		return filename
+
+	def get_filenames(self, distdir):
+		for dirpath, dirnames, filenames in os.walk(distdir,
+				onerror=_raise_exc):
+			return iter(filenames)
+
+	@staticmethod
+	def verify_args(args):
+		return len(args) == 1
+
+
+class FilenameHashLayout(object):
+	def __init__(self, algo, cutoffs):
+		self.algo = algo
+		self.cutoffs = [int(x) for x in cutoffs.split(':')]
+
+	def get_path(self, filename):
+		fnhash = checksum_str(filename.encode('utf8'), self.algo)
+		ret = ''
+		for c in self.cutoffs:
+			assert c % 4 == 0
+			c = c // 4
+			ret += fnhash[:c] + '/'
+			fnhash = fnhash[c:]
+		return ret + filename
+
+	def get_filenames(self, distdir):
+		pattern = ''
+		for c in self.cutoffs:
+			assert c % 4 == 0
+			c = c // 4
+			pattern += c * '[0-9a-f]' + '/'
+		pattern += '*'
+		return (x.rsplit('/', 1)[1]
+				for x in glob(os.path.join(distdir, pattern)))
+
+	@staticmethod
+	def verify_args(args):
+		if len(args) != 3:
+			return False
+		if args[1] not in get_valid_checksum_keys():
+			return False
+		# argsidate cutoffs
+		for c in args[2].split(':'):
+			try:
+				c = int(c)
+			except ValueError:
+				break
+			else:
+				if c % 4 != 0:
+					break
+		else:
+			return True
+		return False
+
+
+class MirrorLayoutConfig(object):
+	"""
+	Class to read layout.conf from a mirror.
+	"""
+
+	def __init__(self):
+		self.structure = ()
+
+	def read_from_file(self, f):
+		cp = SafeConfigParser()
+		read_configs(cp, [f])
+		vals = []
+		for i in itertools.count():
+			try:
+				vals.append(tuple(cp.get('structure', '%d' % i).split()))
+			except ConfigParserError:
+				break
+		self.structure = tuple(vals)
+
+	def serialize(self):
+		return self.structure
+
+	def deserialize(self, data):
+		self.structure = data
+
+	@staticmethod
+	def validate_structure(val):
+		if val[0] == 'flat':
+			return FlatLayout.verify_args(val)
+		if val[0] == 'filename-hash':
+			return FilenameHashLayout.verify_args(val)
+		return False
+
+	def get_best_supported_layout(self):
+		for val in self.structure:
+			if self.validate_structure(val):
+				if val[0] == 'flat':
+					return FlatLayout(*val[1:])
+				elif val[0] == 'filename-hash':
+					return FilenameHashLayout(*val[1:])
+		else:
+			# fallback
+			return FlatLayout()
+
+	def get_all_layouts(self):
+		ret = []
+		for val in self.structure:
+			if not self.validate_structure(val):
+				raise ValueError("Unsupported structure: {}".format(val))
+			if val[0] == 'flat':
+				ret.append(FlatLayout(*val[1:]))
+			elif val[0] == 'filename-hash':
+				ret.append(FilenameHashLayout(*val[1:]))
+		if not ret:
+			ret.append(FlatLayout())
+		return ret
+
+
+def get_mirror_url(mirror_url, filename, cache_path=None):
+	"""
+	Get correct fetch URL for a given file, accounting for mirror
+	layout configuration.
+
+	@param mirror_url: Base URL to the mirror (without '/distfiles')
+	@param filename: Filename to fetch
+	@param cache_path: Path for mirror metadata cache
+	@return: Full URL to fetch
+	"""
+
+	mirror_conf = MirrorLayoutConfig()
+
+	cache = {}
+	if cache_path is not None:
+		try:
+			with open(cache_path, 'r') as f:
+				cache = json.load(f)
+		except (IOError, ValueError):
+			pass
+
+	ts, data = cache.get(mirror_url, (0, None))
+	# refresh at least daily
+	if ts >= time.time() - 86400:
+		mirror_conf.deserialize(data)
+	else:
+		try:
+			f = urlopen(mirror_url + '/distfiles/layout.conf')
+			try:
+				data = io.StringIO(f.read().decode('utf8'))
+			finally:
+				f.close()
+
+			mirror_conf.read_from_file(data)
+		except (ConfigParserError, IOError, UnicodeDecodeError):
+			# Do not cache negative results.
+			cache_path = None
+
+		cache[mirror_url] = (time.time(), mirror_conf.serialize())
+		if cache_path is not None:
+			f = atomic_ofstream(cache_path, 'w')
+			json.dump(cache, f)
+			f.close()
+
+	return (mirror_url + "/distfiles/" +
+			mirror_conf.get_best_supported_layout().get_path(filename))
+
+
 def fetch(myuris, mysettings, listonly=0, fetchonly=0,
 	locks_in_subdir=".locks", use_locks=1, try_mirrors=1, digests=None,
 	allow_missing_digests=True):
@@ -350,11 +527,12 @@ def fetch(myuris, mysettings, listonly=0, fetchonly=0,
 	if listonly or ("distlocks" not in features):
 		use_locks = 0
 
+	distdir_writable = os.access(mysettings["DISTDIR"], os.W_OK)
 	fetch_to_ro = 0
 	if "skiprocheck" in features:
 		fetch_to_ro = 1
 
-	if not os.access(mysettings["DISTDIR"],os.W_OK) and fetch_to_ro:
+	if not distdir_writable and fetch_to_ro:
 		if use_locks:
 			writemsg(colorize("BAD",
 				_("!!! For fetching to a read-only filesystem, "
@@ -434,8 +612,14 @@ def fetch(myuris, mysettings, listonly=0, fetchonly=0,
 	for myfile, myuri in file_uri_tuples:
 		if myfile not in filedict:
 			filedict[myfile]=[]
-			for y in range(0,len(locations)):
-				filedict[myfile].append(locations[y]+"/distfiles/"+myfile)
+			if distdir_writable:
+				mirror_cache = os.path.join(mysettings["DISTDIR"],
+						".mirror-cache.json")
+			else:
+				mirror_cache = None
+			for l in locations:
+				filedict[myfile].append(functools.partial(
+					get_mirror_url, l, myfile, mirror_cache))
 		if myuri is None:
 			continue
 		if myuri[:9]=="mirror://":
@@ -608,7 +792,7 @@ def fetch(myuris, mysettings, listonly=0, fetchonly=0,
 				pruned_digests["size"] = size
 
 		myfile_path = os.path.join(mysettings["DISTDIR"], myfile)
-		download_path = myfile_path + _download_suffix
+		download_path = myfile_path if fetch_to_ro else myfile_path + _download_suffix
 		has_space = True
 		has_space_superuser = True
 		file_lock = None
@@ -876,7 +1060,8 @@ def fetch(myuris, mysettings, listonly=0, fetchonly=0,
 										"File renamed to '%s'\n\n") % \
 										temp_filename, noiselevel=-1)
 							else:
-								_movefile(download_path, myfile_path, mysettings=mysettings)
+								if not fetch_to_ro:
+									_movefile(download_path, myfile_path, mysettings=mysettings)
 								eout = EOutput()
 								eout.quiet = \
 									mysettings.get("PORTAGE_QUIET", None) == "1"
@@ -895,6 +1080,8 @@ def fetch(myuris, mysettings, listonly=0, fetchonly=0,
 			tried_locations = set()
 			while uri_list:
 				loc = uri_list.pop()
+				if isinstance(loc, functools.partial):
+					loc = loc()
 				# Eliminate duplicates here in case we've switched to
 				# "primaryuri" mode on the fly due to a checksum failure.
 				if loc in tried_locations:
@@ -993,7 +1180,7 @@ def fetch(myuris, mysettings, listonly=0, fetchonly=0,
 							del e
 							fetched = 0
 						else:
-							if mystat.st_size < fetch_resume_size:
+							if distdir_writable and mystat.st_size < fetch_resume_size:
 								writemsg(_(">>> Deleting distfile with size "
 									"%d (smaller than " "PORTAGE_FETCH_RESU"
 									"ME_MIN_SIZE)\n") % mystat.st_size)
@@ -1131,13 +1318,14 @@ def fetch(myuris, mysettings, listonly=0, fetchonly=0,
 										(reason[1], reason[2]), noiselevel=-1)
 									if reason[0] == _("Insufficient data for checksum verification"):
 										return 0
-									temp_filename = \
-										_checksum_failure_temp_file(
-											mysettings, mysettings["DISTDIR"],
-											os.path.basename(download_path))
-									writemsg_stdout(_("Refetching... "
-										"File renamed to '%s'\n\n") % \
-										temp_filename, noiselevel=-1)
+									if distdir_writable:
+										temp_filename = \
+											_checksum_failure_temp_file(
+												mysettings, mysettings["DISTDIR"],
+												os.path.basename(download_path))
+										writemsg_stdout(_("Refetching... "
+											"File renamed to '%s'\n\n") % \
+											temp_filename, noiselevel=-1)
 									fetched=0
 									checksum_failure_count += 1
 									if checksum_failure_count == \
@@ -1154,7 +1342,8 @@ def fetch(myuris, mysettings, listonly=0, fetchonly=0,
 										checksum_failure_max_tries:
 										break
 								else:
-									_movefile(download_path, myfile_path, mysettings=mysettings)
+									if not fetch_to_ro:
+										_movefile(download_path, myfile_path, mysettings=mysettings)
 									eout = EOutput()
 									eout.quiet = mysettings.get("PORTAGE_QUIET", None) == "1"
 									if digests:
@@ -1165,7 +1354,8 @@ def fetch(myuris, mysettings, listonly=0, fetchonly=0,
 									break
 					else: # no digests available
 						if not myret:
-							_movefile(download_path, myfile_path, mysettings=mysettings)
+							if not fetch_to_ro:
+								_movefile(download_path, myfile_path, mysettings=mysettings)
 							fetched=2
 							break
 						elif mydigests!=None:
